"""
File: transformations.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from enum import Enum
from pctheory import pitch


class OTO:
    """
    Represents an ordered tone operator (OTO). If used with a twelve-tone row, it is a row operator (RO).
    Objects of this class are subscriptable. [0] is the index of transposition. [1] is whether or not to
    retrograde (0-no or 1-yes). [2] is the multiplier. Multiplication is performed first, then retrograding,
    then transposition. These operators can be used with pcsegs.
    """
    def __init__(self, T=0, R=0, M=0):
        """
        Creates an OTO
        :param T: The index of transposition
        :param R: Whether or not to retrograde
        :param M: The multiplier
        """
        self._oto = (T, R, M)

    def __eq__(self, other):
        return self._oto[0] == other.oto[0] and self._oto[1] == other.oto[1] and self._oto[2] == other.oto[2]

    def __getitem__(self, item):
        return self._oto[item]

    def __hash__(self):
        return self._oto[0] * 1000 + self._oto[1] * 100 + self._oto[2]

    def __ne__(self, other):
        return self._oto[0] != other.oto[0] or self._oto[1] != other.oto[1] or self._oto[2] != other.oto[2]

    def __repr__(self):
        if self._oto[1] and self._oto[2] != 1:
            return f"T{self._oto[0]}RM{self._oto[2]}"
        elif self._oto[2] != 1:
            return f"T{self._oto[0]}M{self._oto[2]}"
        elif self._oto[1]:
            return f"T{self._oto[0]}R"
        else:
            return f"T{self._oto[0]}"

    def __str__(self):
        if self._oto[1] and self._oto[2] != 1:
            return f"T{self._oto[0]}RM{self._oto[2]}"
        elif self._oto[2] != 1:
            return f"T{self._oto[0]}M{self._oto[2]}"
        elif self._oto[1]:
            return f"T{self._oto[0]}R"
        else:
            return f"T{self._oto[0]}"

    @property
    def oto(self):
        """
        Gets the OTO as a tuple. Index 0 is the index of transposition, index 1 is whether or not to retrograde, and
        index 2 is the multiplier.
        :return: The OTO
        """
        return self._oto

    @oto.setter
    def oto(self, value):
        """
        Sets the OTO using a tuple
        :param value: A tuple
        :return:
        """
        self._oto = value

    def transform(self, pcseg: list):
        """
        Transforms a pcseg
        :param pcseg: A pcseg
        :return: The transformed pcseg
        """
        pcseg2 = list()
        if len(pcseg) > 0:
            if type(pcseg[0]) == pitch.PitchClass24:
                for i in range(len(pcseg)):
                    pcseg2.append(pitch.PitchClass24(pcseg[i].pc * self._oto[2] + self._oto[0]))
            else:
                for i in range(len(pcseg)):
                    pcseg2.append(pitch.PitchClass12(pcseg[i].pc * self._oto[2] + self._oto[0]))
        if self._oto[1]:
            pcseg2.reverse()
        return pcseg2


class UTO:
    """
    Represents an unordered tone operator (UTO), which can be used as a twelve-tone operator (TTO)
    or 24-tone operator (24TO). Objects of this class are subscriptable.
    [0] is the index of transposition. [1] is the multiplier. Multiplication is performed first,
    then transposition.
    """
    def __init__(self, T=0, M=1):
        """
        Creates a UTO
        :param T: The index of transposition
        :param M: The index of multiplication
        """
        self._uto = (T, M)

    def __eq__(self, other):
        return self._uto[0] == other.uto[0] and self._uto[1] == other.uto[1]

    def __getitem__(self, item):
        return self._uto[item]

    def __hash__(self):
        return self._uto[0] * 100 + self._uto[1]

    def __ne__(self, other):
        return self._uto[0] != other.uto[0] or self._uto[1] != other.uto[1]

    def __repr__(self):
        if self._uto[1] != 1:
            return f"T{self._uto[0]}M{self._uto[1]}"
        else:
            return f"T{self._uto[0]}"

    def __str__(self):
        if self._uto[1] != 1:
            return f"T{self._uto[0]}M{self._uto[1]}"
        else:
            return f"T{self._uto[0]}"

    @property
    def uto(self):
        """
        Gets the UTO as a list. Index 0 is the index of transposition, and index 1
        is the multiplier.
        :return: The UTO
        """
        return self._uto

    @uto.setter
    def uto(self, value):
        """
        Sets the UTO using a tuple
        :param value: A tuple
        :return:
        """
        self._uto = value

    def cycles(self, num_pcs=12):
        """
        Gets the cycles of the UTO
        :param num_pcs: The number of possible pcs in the system
        :return: The cycles, as a list of lists
        """
        int_list = [i for i in range(num_pcs)]
        cycles = []
        while len(int_list) > 0:
            cycle = [int_list[0]]
            pc = cycle[0]
            pc = (pc * self._uto[1] + self._uto[0]) % num_pcs
            while pc != cycle[0]:
                cycle.append(pc)
                int_list.remove(pc)
                pc = cycle[len(cycle) - 1]
                pc = (pc * self._uto[1] + self._uto[0]) % num_pcs
            cycles.append(cycle)
            del int_list[0]
        return cycles

    def inverse(self, num_pcs=12):
        """
        Gets the inverse of the UTO
        :param num_pcs: The number of possible pcs in the system
        :return: The inverse
        """
        return UTO((-self._uto[1] * self._uto[0]) % num_pcs, self._uto[1])

    def transform(self, item):
        """
        Transforms a pcset, pcseg, or pc
        :param item: A pcset, pcseg, or pc
        :return: The transformed item
        """
        t = type(item)
        if t == pitch.PitchClass12:
            return pitch.PitchClass12(item.pc * self._uto[1] + self._uto[0])
        elif t == pitch.PitchClass24:
            return pitch.PitchClass24(item.pc * self._uto[1] + self._uto[0])
        else:
            new_item = t()
            if t == set:
                for i in item:
                    new_item.add(self.transform(i))
            if t == list:
                for i in item:
                    new_item.append(self.transform(i))
            return new_item


def find_utos(pcset1: set, pcset2: set):
    """
    Finds the UTOS that transform pcset1 into pcset2
    :param pcset1: A pcset
    :param pcset2: A transformed pcset
    :return: A list of UTOS
    """
    utos = get_utos12()
    utos_final = {}
    for u in utos:
        if utos[u].transform(pcset1) == pcset2:
            utos_final[u] = utos[u]
    return utos_final


def get_otos12():
    """
    Gets OTOs
    :return: A list of OTOs
    """
    otos = {}
    for i in range(12):
        otos[f"T{i}"] = OTO(i)
        otos[f"T{i}R"] = OTO(i, 1)
        otos[f"T{i}M11"] = OTO(i, 0, 11)
        otos[f"T{i}RM11"] = OTO(i, 1, 11)
        otos[f"T{i}M5"] = OTO(i, 0, 5)
        otos[f"T{i}RM5"] = OTO(i, 1, 5)
        otos[f"T{i}M7"] = OTO(i, 0, 7)
        otos[f"T{i}RM7"] = OTO(i, 1, 7)
        otos[f"T{i}I"] = otos[f"T{i}M11"]
        otos[f"T{i}RI"] = otos[f"T{i}RM11"]
    return otos


def get_otos24():
    """
    Gets microtonal OTOs
    :return: A list of microtonal OTOs
    """
    otos = {}
    for i in range(24):
        otos[f"T{i}"] = OTO(i)
        otos[f"T{i}R"] = OTO(i, 1)
        otos[f"T{i}M5"] = OTO(i, 0, 5)
        otos[f"T{i}RM5"] = OTO(i, 1, 5)
        otos[f"T{i}M7"] = OTO(i, 0, 7)
        otos[f"T{i}RM7"] = OTO(i, 1, 7)
        otos[f"T{i}M11"] = OTO(i, 0, 11)
        otos[f"T{i}RM11"] = OTO(i, 1, 11)
        otos[f"T{i}M13"] = OTO(i, 0, 13)
        otos[f"T{i}RM13"] = OTO(i, 1, 13)
        otos[f"T{i}M17"] = OTO(i, 0, 17)
        otos[f"T{i}RM17"] = OTO(i, 1, 17)
        otos[f"T{i}M19"] = OTO(i, 0, 19)
        otos[f"T{i}RM19"] = OTO(i, 1, 19)
        otos[f"T{i}M23"] = OTO(i, 0, 23)
        otos[f"T{i}RM23"] = OTO(i, 1, 23)
        otos[f"T{i}I"] = otos[f"T{i}M23"]
        otos[f"T{i}RI"] = otos[f"T{i}RM23"]
    return otos


def get_utos12():
    """
    Gets the twelve-tone UTOs (TTOs)
    :return: A dictionary of UTOs
    """
    utos = {}
    for i in range(12):
        utos[f"T{i}"] = UTO(i, 1)
        utos[f"T{i}M5"] = UTO(i, 5)
        utos[f"T{i}M7"] = UTO(i, 7)
        utos[f"T{i}M11"] = UTO(i, 11)
    return utos


def get_utos24():
    """
    Gets the 24-tone UTOs (24TOs)
    :return: A dictionary of UTOs
    """
    utos = {}
    for i in range(24):
        utos[f"T{i}"] = UTO(i, 1)
        utos[f"T{i}M5"] = UTO(i, 5)
        utos[f"T{i}M7"] = UTO(i, 7)
        utos[f"T{i}M11"] = UTO(i, 11)
        utos[f"T{i}M13"] = UTO(i, 13)
        utos[f"T{i}M17"] = UTO(i, 17)
        utos[f"T{i}M19"] = UTO(i, 19)
        utos[f"T{i}M23"] = UTO(i, 23)
    return utos


def left_multiply_utos(*args, num_pcs=12):
    """
    Left-multiplies a list of UTOs
    :param args: A collection of UTOs (can be one argument as a list, or multiple UTOs separated by commas.
    The highest index is evaluated first, and the lowest index is evaluated last.
    :param num_pcs: The number of pcs in the system
    :return: The result
    """
    utos = args

    # If the user provided a list object
    if len(args) == 1:
        if type(args[0]) == list:
            utos = args[0]

    if len(utos) == 0:
        return None
    elif len(utos) == 1:
        return utos[0]
    else:
        n = utos[len(utos) - 1][0]
        m = utos[len(utos)-1][1]
        for i in range(len(utos)-2, -1, -1):
            tr_n = utos[i][0]
            mul_n = utos[i][1]
            m = m * mul_n + tr_n
            n *= mul_n
        return UTO(n % num_pcs, m % num_pcs)


def make_uto_list(*args):
    """
    Makes a UTO list
    :return: A UTO list
    """
    uto_list = []
    for uto in args:
        uto_list.append(UTO(uto[0], uto[1]))
    return uto_list



"""
File: pcseg.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pctheory import pcset, pitch, tables, transformations
import random


def bip_n(imb_n: list):
    """
    Gets the BIP_n of a list of set-classes generated by IMB_n
    :param imb_n: The IMB_n list
    :return: The BIP_n
    """
    bip = []
    if len(imb_n) > 0:
        for sc in imb_n:
            name = sc.name_forte.split('-')[1]
            name = name.replace("Z", "")
            bip.append(int(name))
        bip.sort()
    return bip


def create_ormap(pcseg: list):
    """
    Creates the ORMAP of a row
    :param pcseg: A row
    :return: The ORMAP
    """
    omap = {}
    t = type(pcseg[0])
    for i in range(len(pcseg)):
        omap[t(pcseg[i].pc)] = i
    return omap


def generate_pcseg12_from_interval_list(interval_list: list, starting_pc=None):
    """
    Generates a pcseg from an interval list
    :param interval_list: The interval list
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A pcseg
    """
    if starting_pc is None:
        random.seed()
        pcseg = [pitch.PitchClass12(random.randrange(12))]
        for i in range(len(interval_list)):
            pcseg.append(pitch.PitchClass12(pcseg[i].pc + interval_list[i]))
        return pcseg
    else:
        pcseg = [pitch.PitchClass12(starting_pc)]
        for i in range(len(interval_list)):
            pcseg.append(pitch.PitchClass12(pcseg[i].pc + interval_list[i]))
        return pcseg


def generate_pcseg24_from_interval_list(interval_list: list, starting_pc=None):
    """
    Generates a pcseg from an interval list
    :param interval_list: The interval list
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A pcseg
    """
    if starting_pc is None:
        random.seed()
        pcseg = [pitch.PitchClass24(random.randrange(24))]
        for i in range(len(interval_list)):
            pcseg.append(pitch.PitchClass24(pcseg[i].pc + interval_list[i]))
        return pcseg
    else:
        pcseg = [pitch.PitchClass24(starting_pc)]
        for i in range(len(interval_list)):
            pcseg.append(pitch.PitchClass24(pcseg[i].pc + interval_list[i]))
        return pcseg


def generate_random_all_interval_row(name_tables=None, starting_pc=None):
    """
    Generates a random all-interval row
    :param name_tables: A dictionary of name tables
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: An all-interval row
    """
    _tables = name_tables if name_tables is not None else tables.create_tables_row()
    random.seed()
    generator = _tables["allIntervalRowGenerators"][random.randrange(len(_tables["allIntervalRowGenerators"]))]
    row = [pitch.PitchClass12(random.randrange(12) if starting_pc is None else starting_pc)]
    for i in range(1, 12):
        row.append(pitch.PitchClass12(row[i - 1].pc + generator[i - 1]))
    return row


def generate_random_pcseg12(length: int, non_duplicative=False, starting_pc=None):
    """
    Generates a random pcseg
    :param length: The length of the pcseg
    :param non_duplicative: Whether or not duplicate pcs may occur (must be True to generate a row)
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A random pcseg
    """
    random.seed()
    pcseg = [pitch.PitchClass12(random.randrange(12) if starting_pc is None else starting_pc)]
    if non_duplicative and 0 < length < 12:
        pcs = [i for i in range(12)]
        del pcs[pcseg[0].pc]
        for i in range(length - 1):
            j = random.randrange(len(pcs))
            pcseg.append(pitch.PitchClass12(pcs[j]))
            del pcs[j]
    elif not non_duplicative and 0 < length:
        for i in range(length - 1):
            pcseg.append(pitch.PitchClass12(random.randrange(12)))
    else:
        raise ValueError("Invalid length")
    return pcseg


def generate_random_pcseg24(length: int, non_duplicative=False, starting_pc=None):
    """
    Generates a random pcseg
    :param length: The length of the pcseg
    :param non_duplicative: Whether or not duplicate pcs may occur (must be True to generate a row)
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A random pcseg
    """
    random.seed()
    pcseg = [pitch.PitchClass24(random.randrange(24) if starting_pc is None else starting_pc)]
    if non_duplicative and 0 < length < 24:
        pcs = [i for i in range(24)]
        del pcs[pcseg[0].pc]
        for i in range(length - 1):
            j = random.randrange(len(pcs))
            pcseg.append(pitch.PitchClass24(pcs[j]))
            del pcs[j]
    elif not non_duplicative and 0 < length:
        for i in range(length - 1):
            pcseg.append(pitch.PitchClass24(random.randrange(24)))
    else:
        raise ValueError("Invalid length")
    return pcseg


def generate_random_pcseg_from_pcset(pcset: set):
    """
    Generates a random pcseg from a pcset
    :param pcset: A pcset
    :return: A pcseg
    """
    random.seed()
    pcseg = []
    setseg = list(pcset)
    t = type(setseg[0])
    for i in range(len(setseg)):
        j = random.randrange(len(setseg))
        pcseg.append(t(setseg[j].pc))
        del setseg[j]
    return pcseg


def get_intervals(pcseg: list):
    """
    Gets the interval sequence of a pcseg
    :param pcseg: The pcseg
    :return: The interval sequence
    """
    intervals = []
    for i in range(1, len(pcseg)):
        intervals.append(pcseg[i].pc - pcseg[i - 1].pc)
    return intervals


def get_secondary_forms(pcseg: list, subseg: list):
    """
    Gets all secondary forms that contain the provided ordered subseg
    :param pcseg: A pcseg
    :param subseg: A subseg
    :return: A list of secondary forms that contain the subseg, as well as their RO transformations
    """
    secondary_forms = []

    # Get different transformations
    m5 = multiply(pcseg, 5)
    m7 = multiply(pcseg, 7)
    m11 = multiply(pcseg, 11)
    r = retrograde(pcseg)
    rm5 = retrograde(multiply(pcseg, 5))
    rm7 = retrograde(multiply(pcseg, 7))
    rm11 = retrograde(multiply(pcseg, 11))

    # Exhaust all transformations and search them
    for i in range(12):
        t = [[transformations.OTO(i, 0, 1), transpose(pcseg, i)], [transformations.OTO(i, 0, 5), transpose(m5, i)],
             [transformations.OTO(i, 0, 7), transpose(m7, i)], [transformations.OTO(i, 0, 11), transpose(m11, i)],
             [transformations.OTO(i, 1, 1), transpose(r, i)], [transformations.OTO(i, 1, 5), transpose(rm5, i)],
             [transformations.OTO(i, 1, 7), transpose(rm7, i)], [transformations.OTO(i, 1, 11), transpose(rm11, i)]]

        # Search each transformation in t
        for item in t:
            is_valid = False
            for i in range(len(item[1]) - len(subseg)):
                for j in range(i, i + len(subseg)):
                    if item[1][j] == subseg[j - i]:
                        is_valid = True
                    if not is_valid:
                        break
                if is_valid:
                    secondary_forms.append(item)
                    break

    return secondary_forms


def invert(pcseg: list):
    """
    Inverts a pcseg
    :param pcseg: The pcseg
    :return: The inverted pcseg
    """
    pcseg2 = []
    if len(pcseg) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg)))
        for pc in pcseg:
            pcseg2.append(t(pc.pc * -1))
    return pcseg2


def imb_n(pcseg: list, n: int, name_tables=None):
    """
    Gets the IMB_n of a pcseg. The IMB_n is the segment of imbricated set-classes of cardinality n.
    :param pcseg: The pcseg
    :param n: The cardinality of imbrication
    :param name_tables: Name tables
    :return: The IMB_n
    """
    imb = set()
    scs = []
    if type(pcseg[0]) == pitch.PitchClass12:
        if name_tables is None:
            name_tables = tables.create_tables_set()
        for i in range(len(pcseg) + 1 - n):
            for j in range(i, i + n):
                imb.add(pcseg[j])
            scs.append(pcset.SetClass12(name_tables, imb))
            imb.clear()
    elif type(pcseg[0]) == pitch.PitchClass24:
        for i in range(len(pcseg) + 1 - n):
            for j in range(i, i + n):
                imb.add(pcseg[j])
            scs.append(pcset.SetClass24(imb))
            imb.clear()
    return scs


def is_row(pcseg: list):
    """
    Determines if a pcseg is a row
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is a row
    """
    if len(pcseg) != 12:
        return False
    pcset = set([pc.pc for pc in pcseg])
    if len(pcset) < 12:
        return False
    return True


def is_row_generator(rgen: list):
    """
    Determines if a row generator is valid
    :param rgen: A row generator
    :return: True if the row generator is valid; false otherwise
    """
    for i in range(len(list) - 1):
        rgen_sum = rgen[i]
        for j in range(i + 1, len(list)):
            rgen_sum += rgen[j]
            if rgen_sum % 12 == 0:
                return False
    return True


def make_pcseg12(*args):
    """
    Makes a pcseg
    :param *args: Pcs
    :return: A pcseg
    """
    if type(args[0]) == list:
        args = args[0]
    return [pitch.PitchClass12(pc) for pc in args]


def make_pcseg24(*args):
    """
    Makes a pcseg
    :param *args: Pcs
    :return: A pcseg
    """
    if type(args[0]) == list:
        args = args[0]
    return [pitch.PitchClass24(pc) for pc in args]


def multiply(pcseg: list, n: int):
    """
    Multiplies a pcseg
    :param pcseg: The pcseg
    :param n: The multiplier
    :return: The multiplied pcseg
    """
    pcseg2 = []
    if len(pcseg) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg)))
        for pc in pcseg:
            pcseg2.append(t(pc.pc * n))
    return pcseg2


def multiply_order(pcseg: list, n: int):
    """
    Multiplies a pcseg's order
    :param pcseg: The pcseg
    :param n: The multiplier
    :return: The order-multiplied pcseg
    """
    pcseg2 = []
    for i in range(len(pcseg)):
        pcseg2.append(pitch.PitchClass12(pcseg[(i * n) % len(pcseg)].pc))
    return pcseg2


def ormap(row: list, ormap: dict):
    """
    Performs the ORMAP mapping on a row, given a provided ORMAP
    :param row: A row
    :param ormap:
    :return: The ORMAP mapping for the row
    """
    mapping = []
    for item in row:
        mapping.append(ormap[item])
    return mapping


def prot(pcseg: list):
    """
    Generates the protocol pairs for a pcseg
    :param pcseg: A pcseg
    :return: A set of protocol pairs
    """
    pp = set()
    for i in range(len(pcseg)):
        for j in range(i, len(pcseg)):
            pp.add((pcseg[i], pcseg[j]))
    return pp


def retrograde(pcseg: list):
    """
    Retrogrades a pcseg
    :param pcseg: The pcseg
    :return: The retrograded pcseg
    """
    pcseg2 = []
    if len(pcseg) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg)))
        for i in range(len(pcseg) - 1, -1, -1):
            pcseg2.append(t(pcseg[i].pc))
    return pcseg2


def rotate(pcseg: list, n: int):
    """
    Rotates a pcseg
    :param pcseg: The pcseg
    :param n: The index of rotation
    :return: The rotated pcseg
    """
    pcseg2 = []
    if len(pcseg) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg)))
        for i in range(len(pcseg)):
            pcseg2.append(t(pcseg[(i - n) % len(pcseg)].pc))
    return pcseg2


def transpose(pcseg: list, n: int):
    """
    Transposes a pcseg
    :param pcseg: The pcseg
    :param n: The index of transposition
    :return: The transposed pcseg
    """
    pcseg2 = []
    if len(pcseg) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg)))
        for pc in pcseg:
            pcseg2.append(t(pc.pc + n))
    return pcseg2


def transpositional_combination(pcseg1: list, pcseg2: list):
    """
    Transpositionally combines (TC) two pcsegs. This is Boulez's "multiplication."
    :param pcseg1: A pcseg
    :param pcseg2: A pcseg
    :return: The TC pcset
    """
    pcseg3 = []
    if len(pcseg1) > 0 and len(pcseg2) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcseg1)))
        for pc2 in pcseg2:
            for pc1 in pcseg1:
                pcseg3.append(t(pc1.pc + pc2.pc))
    return pcseg3


class InvarianceMatrix:
    """
    Represents an invariance matrix
    """
    def __init__(self, mx_type="T", a=None, c=None):
        """
        Creates an invariance matrix
        :param mx_type: The matrix type (T, I, M, or MI)
        :param a: Pcseg A
        :param c: Pcseg B
        """
        self._a = None
        self._b = None
        self._c = None
        self._mx_type = mx_type.upper()
        self._mx = None
        self._t = None
        if a is not None and c is not None:
            self.load_matrix(a, c)

    def __getitem__(self, i, j):
        """
        Gets the pc at the specified row and column
        :param i: The row
        :param j: The column
        :return: The pc
        """
        return self._mx[i][j]

    def __repr__(self):
        """
        Gets a representation of the InvarianceMatrix object
        :returns: A string representation of the InvarianceMatrix object
        """
        return "<pctheory.pcseg.InvarianceMatrix object at " + str(id(self)) + ">: " + str(self._mx)

    def __str__(self):
        """
        Converts the InvarianceMatrix object to string
        :returns: A string conversion of the InvarianceMatrix object
        """
        chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B']
        lines = "    "
        for pc in self._a:
            lines += chars[pc.pc] + " "
        lines += "\n   "
        for i in range(2 * len(self._a)):
            lines += "-"
        lines += "\n"
        for i in range(len(self._mx) - 1):
            lines += chars[self._b[i].pc] + " | "
            for j in range(len(self._mx[i])):
                lines += chars[self._mx[i][j].pc] + " "
            lines += "\n"
        lines += chars[self._b[len(self._mx) - 1].pc] + " | "
        for j in range(len(self._mx[len(self._mx) - 1])):
            lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
        return lines

    @property
    def mx(self):
        """
        Gets the invariance matrix
        :return: The invariance matrix
        """
        return self._mx

    @property
    def pcseg_a(self):
        """
        Gets the pcseg A
        :return: The pcseg A
        """
        return self._a

    @property
    def pcseg_b(self):
        """
        Gets the pcseg B
        :return: The pcseg B
        """
        return self._b

    @property
    def pcseg_c(self):
        """
        Gets the pcseg C
        :return: The pcseg C
        """
        return self._c

    def get_column(self, j):
        """
        Gets a column of the matrix
        :param j: The column index
        :return: The column
        """
        c = []
        for i in range(len(self._mx)):
            c.append(self._mx[i][j])
        return c

    def get_row(self, i):
        """
        Gets a row of the matrix
        :param i: The row index
        :return: The row
        """
        return self._mx[i]

    def load_matrix(self, a: list, c: list):
        """
        Loads the matrix
        :param a: Pcseg A
        :param c: Pcseg C
        :return: None
        """
        self._t = type(a[0])
        ro = transformations.OTO()
        if self._t == pitch.PitchClass12:
            INVERT = 11
            if self._mx_type == "T":
                ro.oto = [0, 0, 1 * INVERT % 12]
            elif self._mx_type == "M" or self._mx_type == "M5":
                ro.oto = [0, 0, 5 * INVERT % 12]
            elif self._mx_type == "MI" or self._mx_type == "M7":
                ro.oto = [0, 0, 7 * INVERT % 12]
            elif self._mx_type == "I" or self._mx_type == "M11":
                ro.oto = [0, 0, 11 * INVERT % 12]
            b = ro.transform(c)
        if self._t == pitch.PitchClass24:
            INVERT = 23
            if self._mx_type == "T":
                ro.oto = [0, 0, 1 * INVERT % 24]
            elif self._mx_type == "M5":
                ro.oto = [0, 0, 5 * INVERT % 24]
            elif self._mx_type == "M7":
                ro.oto = [0, 0, 7 * INVERT % 24]
            elif self._mx_type == "M11":
                ro.oto = [0, 0, 11 * INVERT % 24]
            elif self._mx_type == "M13":
                ro.oto = [0, 0, 13 * INVERT % 24]
            elif self._mx_type == "M17":
                ro.oto = [0, 0, 17 * INVERT % 24]
            elif self._mx_type == "M19":
                ro.oto = [0, 0, 19 * INVERT % 24]
            elif self._mx_type == "I" or self._mx_type == "M23":
                ro.oto = [0, 0, 23 * INVERT % 24]
            b = ro.transform(c)
        self._mx = []
        for i in range(len(b)):
            mxrow = []
            for j in range(len(a)):
                mxrow.append(self._t(b[i].pc + a[j].pc))
            self._mx.append(mxrow)
        self._a = a.copy()
        self._b = b
        self._c = c.copy()

    def print(self, include: list = None):
        """
        Prints the invariance matrix
        :param include: The pcs to include (if None, all pcs will be printed)
        :returns: None
        """
        chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B']
        lines = "    "
        for pc in self._a:
            lines += chars[pc.pc] + " "
        lines += "\n   "
        for i in range(2 * len(self._a)):
            lines += "-"
        lines += "\n"
        if include is None:
            for i in range(len(self._mx) - 1):
                lines += chars[self._b[i].pc] + " | "
                for j in range(len(self._mx[i])):
                    lines += chars[self._mx[i][j].pc] + " "
                lines += "\n"
            lines += chars[self._b[len(self._mx) - 1].pc] + " | "
            for j in range(len(self._mx[len(self._mx) - 1])):
                lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
        else:
            for i in range(len(self._mx) - 1):
                lines += chars[self._b[i].pc] + " | "
                for j in range(len(self._mx[i])):
                    if self._mx[i][j] in include:
                        lines += chars[self._mx[i][j].pc] + " "
                    else:
                        lines += "  "
                lines += "\n"
            lines += chars[self._b[len(self._mx) - 1].pc] + " | "
            for j in range(len(self._mx[len(self._mx) - 1])):
                if self._mx[len(self._mx) - 1][j] in include:
                    lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
                else:
                    lines += "  "
        print(lines)


class TwelveToneMatrix:
    """
    Represents a twelve-tone matrix
    """
    def __init__(self, row=None):
        """
        Creates a twelve-tone matrix
        :param row: A row to import
        """
        self._labels_left = []
        self._labels_right = []
        self._labels_top = []
        self._labels_bottom = []
        self._mx = []
        self._row = None
        if row is not None:
            self.import_row(row)

    def __getitem__(self, i, j):
        """
        Gets the pc at the specified row and column
        :param i: The row
        :param j: The column
        :return: The pc
        """
        return self._mx[i][j]

    def __repr__(self):
        """
        Generates a string representation of the TwelveToneMatrix that can be printed
        :return: A string representation of the TwelveToneMatrix
        """
        lines = "     "
        for i in range(12):
            lines += f"{self._labels_top[i]: <5} "
        lines += "\n"
        for i in range(12):
            lines += f"{self._labels_left[i]: <4}"
            for j in range(12):
                lines += f"  {str(self._mx[i][j])}   "
            lines += f"{self._labels_right[i]: >4}"
            lines += "\n"
        lines += "     "
        for i in range(12):
            lines += f"{self._labels_bottom[i]: <5} "
        return lines

    def __str__(self):
        """
        Generates a string representation of the TwelveToneMatrix that can be printed
        :return: A string representation of the TwelveToneMatrix
        """
        lines = "     "
        for i in range(12):
            lines += f"{self._labels_top[i]: <5} "
        lines += "\n"
        for i in range(12):
            lines += f"{self._labels_left[i]: <4}"
            for j in range(12):
                lines += f"  {str(self._mx[i][j])}   "
            lines += f"{self._labels_right[i]: >4}"
            lines += "\n"
        lines += "     "
        for i in range(12):
            lines += f"{self._labels_bottom[i]: <5} "
        return lines

    @property
    def labels_bottom(self):
        """
        Gets the bottom matrix labels
        :return: The bottom matrix labels
        """
        return self._labels_bottom

    @property
    def labels_left(self):
        """
        Gets the left matrix labels
        :return: The left matrix labels
        """
        return self._labels_left

    @property
    def labels_right(self):
        """
        Gets the right matrix labels
        :return: The right matrix labels
        """
        return self._labels_right

    @property
    def labels_top(self):
        """
        Gets the top matrix labels
        :return: The top matrix labels
        """
        return self._labels_top

    @property
    def matrix(self):
        """
        Gets the matrix
        :return: The matrix
        """
        return self._mx

    @property
    def row(self):
        """
        Gets the row
        :return: The row
        """
        return self._row

    def get_column(self, j):
        """
        Gets a column of the matrix
        :param j: The column index
        :return: The column
        """
        pcseg = []
        for i in range(len(self._mx)):
            pcseg.append(pitch.PitchClass12(self._mx[i][j].pc))
        return pcseg

    def get_row(self, i):
        """
        Gets a row of the matrix
        :param i: The row index
        :return: The row
        """
        return self._mx[i].copy()

    def import_row(self, row: list):
        """
        Imports a row
        :param row: The row to import
        :return:
        """
        # We need the starting pitch to be 0 for the P0 form
        self._row = transpose(row, -row[0].pc)
        inv = invert(self._row)

        for i in range(len(self._row)):
            self._mx.append(transpose(self._row, inv[i].pc))

        for i in range(len(self._row)):
            p_lbl = self._mx[i][0] - self._mx[0][0]
            i_lbl = self._mx[0][i] - self._mx[0][0]
            r_lbl = self._mx[i][len(self._row) - 1] - self._mx[0][len(self._row) - 1]
            ri_lbl = self._mx[len(self._row) - 1][i] - self._mx[len(self._row) - 1][0]
            self._labels_bottom.append(f"T{ri_lbl.pc}RI")
            self._labels_left.append(f"T{p_lbl.pc}")
            self._labels_right.append(f"T{r_lbl.pc}R")
            self._labels_top.append(f"T{i_lbl.pc}I")



"""
File: array.py
Author: Jeff Martin
Date: 2/2/2022

Copyright © 2022 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pctheory import pcset, poset, pitch, tables, transformations
from pctheory.pcseg import rotate, transpose


class RotationalArray:
    """
    Represents a rotational array
    """
    def __init__(self, pcseg=None):
        """
        Creates a rotational array
        :param pcseg: A pcseg to import
        """
        self._array = None
        self._pcseg = None
        if pcseg is not None:
            self.import_pcseg(pcseg)

    def __repr__(self):
        return "<pctheory.pcseg.RotationalArray object at " + str(id(self)) + ">: " + str(self._array)

    def __str__(self):
        chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B']
        lines = ""
        for i in range(len(self._array) - 1):
            for j in range(len(self._array[i])):
                lines += chars[self._array[i][j].pc] + " "
            lines += "\n"
        for j in range(len(self._array[len(self._array) - 1])):
            lines += chars[self._array[len(self._array) - 1][j].pc] + " "
        return lines

    @property
    def array(self):
        """
        Gets the rotational array
        :return: The rotational array
        """
        return self._array

    @property
    def pcseg(self):
        """
        Gets the pcseg
        :return: The pcseg
        """
        return self._pcseg

    def __getitem__(self, i: int, j: int):
        """
        Gets the pc at the specified row and column
        :param i: The row
        :param j: The column
        :return: The pc
        """
        return self._array[i][j]

    def get_column(self, j: int):
        """
        Gets a column of the rotational array
        :param j: The column index
        :return: The column
        """
        c = []
        for i in range(len(self._array)):
            c.append(self._array[i][j])
        return c

    def get_row(self, i: int):
        """
        Gets a row of the rotational array
        :param i: The row index
        :return: The row
        """
        return self._array[i]

    def import_pcseg(self, pcseg: list):
        """
        Imports a pcseg
        :param pcseg: A pcseg
        :return: None
        """
        self._pcseg = transpose(pcseg, 12 - pcseg[0].pc)
        self._array = []
        for i in range(len(self._pcseg)):
            self._array.append(rotate(transpose(self._pcseg, -self._pcseg[i].pc), len(self._pcseg) - i))


def str_array(array):
    """
    Converts an array of pcs to string
    :param array: The array to convert
    :return: The string
    """
    str_temp = ""
    if type(array) == list:
        for l in array:
            for pc in l:
                str_temp += f"{pc} "
            str_temp += "\n"
    elif type(array) == np.array:
        for i in range(array.shape[0]):
            for j in range(array.shape[1]):
                str_temp += f"{array[i][j]} "
            str_temp += "\n"
    return str_temp


def make_array_chain(array, length: int):
    """
    Makes a chain of arrays
    :param array: An array
    :param length: The length
    :param alt_ret: Whether or not to alternately retrograde the arrays.
    :return: A chained array
    """
    array1 = []  # The final array
    pcset_start = set()  # The begin-set of the array
    pcset_end = set()  # The end-set of the array

    # Populate the begin-set and end-set
    for i in range(len(array)):
        pcset_start.add(array[i][0])
        pcset_end.add(array[i][len(array[i]) - 1])

    # Add the first array to the final array
    for i in range(len(array)):
        row = []
        for j in range(len(array[i])):
            if type(array[i][j]) == list:
                row.append(list(array[i][j]))
            elif type(array[i][j]) == set:
                row.append(set(array[i][j]))
            else:
                row.append(array[i][j])
        array1.append(row)

    # Add the other arrays
    for i in range(1, length):
        # Get the current end-set
        pcset_end_temp = set()
        for j in range(len(array1)):
            pcset_end_temp.add(array1[j][len(array1[j]) - 1])

        # Get the row operator we need for the transformation, and transform the array
        r = transformations.OTO()
        transformation = transformations.find_utos(pcset_start, pcset_end_temp)
        r.oto = [transformation[0][0], 0, transformation[0][1], transformation[0][2]]
        m = transform_row_content(array1, r)
        m.reverse()

        # Add the transformed array content to the end of the large array
        for j in range(len(array1)):
            for k in range(len(m)):
                if m[k][0] == array1[j][len(array1[j]) - 1]:
                    for n in range(1, len(m[k])):
                        array1[j].append(m[k][n])
                    del m[k]
                    break

    return array1



"""
File: pcset.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import networkx
import pyvis

from pctheory import pitch, tables, transformations


class SetClass12:
    """
    Represents a pc-set-class
    """
    NUM_PC = 12

    def __init__(self, name_tables=None, pcset=None):
        """
        Creates a SetClass
        :param name_tables: A list of name tables
        :param pcset: A pcset to initialize the SetClass
        """
        self._dsym = SetClass12.NUM_PC * 2
        self._ic_vector = [0 for i in range(SetClass12.NUM_PC // 2)]
        self._ic_vector_long = [0 for i in range(SetClass12.NUM_PC // 2 + 1)]
        self._name_carter = ""
        self._name_forte = ""
        self._name_morris = ""
        self._name_prime = ""
        self._num_forte = 0
        self._pcset = set()
        self._weight_right = True
        self._tables = name_tables if name_tables is not None else tables.create_tables_set()
        if pcset is not None:
            self.pcset = pcset

    def __eq__(self, other):
        return self.pcset == other.pcset

    def __hash__(self):
        return len(self) * 100 + self.num_forte

    def __len__(self):
        return len(self._pcset)

    def __lt__(self, other):
        if len(self) < len(other) or (len(self) == len(other) and self.num_forte < other.num_forte):
            return True
        return False

    def __ne__(self, other):
        return self.pcset != other.pcset

    def __repr__(self):
        # return "<pctheory.pcset.SetClass12 object at " + str(id(self)) + ">: " + repr(self._pcset)
        return self._name_morris

    def __str__(self):
        return str([str(pc) for pc in self._pcset])

    @property
    def derived_core(self):
        """
        Gets derived core associations
        :return: The derived core associations (or None if not derived core)
        """
        if self.name_prime in self._tables["carterDerivedCoreTable"]:
            return [name for name in self._tables["carterDerivedCoreTable"][self.name_prime]]
        else:
            return None

    @property
    def dsym(self):
        """
        Gets the degree of symmetry of the set-class.
        :return: The degree of symmetry
        """
        return self._dsym

    @property
    def ic_vector(self):
        """
        Gets the IC vector
        :return: The IC vector
        """
        return self._ic_vector

    @property
    def ic_vector_long(self):
        """
        Gets the IC vector in long format
        :return: The IC vector in long format
        """
        return self._ic_vector_long

    @property
    def ic_vector_str(self):
        """
        Gets the IC vector as a string
        :return: The IC vector
        """
        s = "["
        for a in self._ic_vector:
            s += self._tables["hexChars"][a]
        s += "]"
        return s

    @property
    def ic_vector_long_str(self):
        """
        Gets the IC vector in long format as a string
        :return: The IC vector in long format
        """
        s = "["
        for a in self._ic_vector_long:
            s += self._tables["hexChars"][a]
        s += "]"
        return s

    @property
    def is_z_relation(self):
        """
        Whether or not this set-class is Z-related to another set-class
        :return: A boolean
        """
        if "Z" in self.name_forte:
            return True
        return False

    @property
    def name_carter(self):
        """
        Generates the Carter name for a set-class
        :return: The Carter name
        """
        return self._name_carter

    @property
    def name_forte(self):
        """
        Generates the Forte name for a set-class
        :return: The Forte name
        """
        return self._name_forte

    @property
    def name_morris(self):
        """
        Generates the Morris name for a set-class
        :return: The Morris name
        """
        return self._name_morris

    @property
    def name_prime(self):
        """
        Generates the prime-form name (Rahn) for a set-class
        :return: The prime-form name
        """
        return self._name_prime

    @property
    def num_forte(self):
        """
        Generates the number part of the Forte name
        :return: The number part of the Forte name
        """
        return self._num_forte

    @property
    def pcset(self):
        """
        Gets the pcset prime form
        :return: The pcset prime form
        """
        return self._pcset

    @pcset.setter
    def pcset(self, value):
        """
        Updates the pcset prime form
        :param value: The new pcset
        :return:
        """
        self._pcset = SetClass12.calculate_prime_form(value, self._weight_right)
        self._make_names()

    @property
    def weight_right(self):
        """
        Whether or not to weight from the right
        :return: A Boolean
        """
        return self._weight_right

    @weight_right.setter
    def weight_right(self, value: bool):
        """
        Whether or not to weight from the right
        :param value: A Boolean
        :return:
        """
        self._weight_right = value
        self._pcset = SetClass12.calculate_prime_form(self._pcset, self._weight_right)

    @staticmethod
    def calculate_prime_form(pcset: set, weight_from_right: bool = True):
        """
        Calculates the prime form of a pcset
        :param pcset: The pcset
        :param weight_from_right: Whether or not to pack from the right
        :return: The prime form
        """
        prime_set = set()
        if len(pcset) > 0:
            lists_to_weight = []
            int_set = SetClass12._make_int_set(pcset)
            pclist = list(int_set)
            inverted = list(SetClass12._make_int_set(set(SetClass12._invert(pclist))))
            prime_list = None

            # Add regular forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i].append(pclist[i2])
                for i2 in range(0, i):
                    lists_to_weight[i].append(pclist[i2])
                initial_pitch = lists_to_weight[i][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i][i2] -= initial_pitch
                    if lists_to_weight[i][i2] < 0:
                        lists_to_weight[i][i2] += SetClass12.NUM_PC
                lists_to_weight[i].sort()

            # Add inverted forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                for i2 in range(0, i):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                initial_pitch = lists_to_weight[i + len(pclist)][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i + len(pclist)][i2] -= initial_pitch
                    if lists_to_weight[i + len(pclist)][i2] < 0:
                        lists_to_weight[i + len(pclist)][i2] += SetClass12.NUM_PC
                lists_to_weight[i + len(pclist)].sort()

            # Weight lists
            if weight_from_right:
                prime_list = SetClass12._weight_from_right(lists_to_weight)
            else:
                prime_list = SetClass12._weight_left(lists_to_weight)

            # Create pcset
            for pc in prime_list:
                prime_set.add(pitch.PitchClass12(pc))

        return prime_set

    def contains_abstract_subset(self, sc):
        """
        Determines if a set-class is an abstract subset of this set-class
        :param sc: A set-class
        :return: A boolean
        """
        tr = []
        inv = invert(sc.pcset)
        for i in range(SetClass12.NUM_PC):
            tr.append(transpose(sc.pcset, i))
            tr.append(transpose(inv, i))
        for pcs in tr:
            if pcs.issubset(self.pcset):
                return True
        return False

    def get_abstract_complement(self):
        """
        Gets the abstract complement of the SetClass
        :return: The abstract complement SetClass
        """
        csc = SetClass12(self._tables)
        csc.pcset = get_complement(self._pcset)
        return csc

    def get_invariance_vector(self):
        """
        Gets the invariance vector of the SetClass
        :return: The invariance vector
        """
        iv = [0, 0, 0, 0, 0, 0, 0, 0]
        c = get_complement(self._pcset)
        utos = transformations.get_utos12()
        for i in range(SetClass12.NUM_PC):
            h = [utos[f"T{i}"].transform(self._pcset), utos[f"T{i}I"].transform(self._pcset),
                 utos[f"T{i}M"].transform(self._pcset), utos[f"T{i}MI"].transform(self._pcset)]
            for j in range(4):
                if h[j] == self._pcset:
                    iv[j] += 1
                if h[j].issubset(c):
                    iv[4 + j] += 1
        return iv

    def get_abstract_subset_classes(self):
        """
        Gets a set of subset-classes contained in this SetClass
        :return:
        """
        sub = subsets(self._pcset)
        subset_classes = set()
        for s in sub:
            subset_classes.add(SetClass12(self._tables, s))
        return subset_classes

    def get_partition2_subset_classes(self):
        """
        Gets a set of set-class partitions of this SetClass
        :return:
        """
        p = partitions2(self._pcset)
        p_set = set()
        for part in p:
            p_set.add((SetClass12(self._tables, part[0]), SetClass12(self._tables, part[1])))
        return p_set

    def get_z_relation(self):
        """
        Gets the Z-relation of the SetClass
        :return: The Z-relation of the SetClass
        """
        zset = SetClass12(self._tables)
        f = self.name_forte
        if "Z" in f:
            zset.load_from_name(self._tables["zNameTable"][f])
        return zset

    def is_valid_name(self, name: str):
        """
        Determines if a set-class name is valid. Validates prime form, Forte, and Morris names.
        Prime form name format: [xxxx]
        Forte name format: x-x
        Morris name format: (x-x)[xxxx]
        :param name: The name
        :return: A boolean
        """
        if "[" in name and "-" in name:
            name = name.split(")")
            name[0] = name[0].replace("(", "")
            if name[0] in self._tables["forteToSetNameTable"] and name[1] in self._tables["setToForteNameTable"]:
                return True
        elif "-" in name:
            if name in self._tables["forteToSetNameTable"]:
                return True
        elif name in self._tables["setToForteNameTable"] or name in self._tables["setToForteNameTableLeftPacking"]:
            return True
        return False

    def load_from_name(self, name: str):
        """
        Loads a set-class from a prime-form, Morris, or Forte name
        :param name: The name
        :return:
        """
        valid = True
        pname = ""
        if "[" in name and "-" in name:
            pname = name.split("[")[1]
        elif "-" in name:
            # Allow Forte names without Z
            name2 = name.split('-')
            name2 = "-Z".join(name2)
            if name in self._tables["forteToSetNameTable"]:
                pname = self._tables["forteToSetNameTable"][name]
            elif name2 in self._tables["forteToSetNameTable"]:
                pname = self._tables["forteToSetNameTable"][name2]
            else:
                valid = False
        elif name in self._tables["setToForteNameTable"]:
            pname = name
        else:
            valid = False
        if valid:
            pname = pname.replace("[", "")
            pname = pname.replace("]", "")
            pname = [c for c in pname]
            pcset = set([pitch.PitchClass12(self._tables["hexToInt"][pn]) for pn in pname])
            self.pcset = pcset

    @staticmethod
    def _invert(pcseg: list):
        """
        Inverts a pcseg
        :param pcset: The pcseg
        :return: The inverted pcseg
        """
        pcseg2 = []
        for pc in pcseg:
            pcseg2.append(pitch.PitchClass12((pc * 11) % SetClass12.NUM_PC))
        return pcseg2

    @staticmethod
    def _make_int_set(pcset: set):
        """
        Makes an int set
        :param pcset: A pcset
        :return: An int set
        """
        pcset2 = set()
        for pc in pcset:
            pcset2.add(pc.pc)
        return pcset2

    def _make_names(self):
        """
        Makes the names for the set-class
        :return:
        """
        pc_name_list = [self._tables["hexChars"][pc.pc] for pc in self._pcset]
        pc_name_list.sort()
        self._name_prime = "[" + "".join(pc_name_list) + "]"
        if self._name_prime != "[]":
            if self._name_prime in self._tables["setToForteNameTableLeftPacking"]:
                self._name_forte = self._tables["setToForteNameTableLeftPacking"][self._name_prime]
            else:
                self._name_forte = self._tables["setToForteNameTable"][self._name_prime]
        else:
            self._name_forte = "0-1"
        self._name_carter = ""
        if self._name_forte in self._tables["forteToCarterNameTable"]:
            self._name_carter = self._tables["forteToCarterNameTable"][self._name_forte]
        self._name_morris = "(" + self._name_forte + ")" + self._name_prime
        forte_num = self.name_forte.split('-')[1]
        forte_num = forte_num.strip('Z')
        self._num_forte = int(forte_num)
        self._ic_vector_long = [0 for i in range(SetClass12.NUM_PC // 2 + 1)]
        for pc in self._pcset:
            for pc2 in self._pcset:
                interval = (pc2.pc - pc.pc) % SetClass12.NUM_PC
                if interval > SetClass12.NUM_PC // 2:
                    interval = interval * -1 + SetClass12.NUM_PC
                self._ic_vector_long[interval] += 1
        for i in range(1, len(self._ic_vector_long)):
            self._ic_vector_long[i] //= 2
        self._ic_vector = self._ic_vector_long[1:]
        if len(self._pcset) > 0:
            c = get_corpus(self._pcset)
            self._dsym = (SetClass12.NUM_PC * 2) // len(c)
        else:
            self._dsym = SetClass12.NUM_PC * 2

    @staticmethod
    def _weight_from_right(pclists: list):
        """
        Weights pclists from the right
        :param pclists: Pclists
        :return: The most weighted form
        """
        for i in range(len(pclists[0]) - 1, -1, -1):
            if len(pclists) > 1:
                # The smallest item at the current index
                smallest_item = SetClass12.NUM_PC - 1

                # Identify the smallest item at the current index
                for j in range(len(pclists)):
                    if pclists[j][i] < smallest_item:
                        smallest_item = pclists[j][i]

                # Remove all lists with larger items at the current index
                j = 0
                while j < len(pclists):
                    if pclists[j][i] > smallest_item:
                        del pclists[j]
                    else:
                        j += 1

            else:
                break
        return pclists[0]

    @staticmethod
    def _weight_left(pclists: list):
        """
        Weights pclists left
        :param pclists: Pclists
        :return: The most weighted form
        """
        if len(pclists) > 1:
            # The smallest item at the current index
            smallest_item = SetClass12.NUM_PC - 1

            # Identify the smallest item at the last index
            for j in range(0, len(pclists)):
                if pclists[j][len(pclists[0]) - 1] < smallest_item:
                    smallest_item = pclists[j][len(pclists[0]) - 1]

            # Remove all lists with larger items at the current index
            j = 0
            while j < len(pclists):
                if pclists[j][len(pclists[0]) - 1] > smallest_item:
                    del pclists[j]
                else:
                    j += 1

            # Continue processing, but now pack from the left
            for i in range(0, len(pclists[0])):
                if len(pclists) > 1:
                    smallest_item = SetClass12.NUM_PC - 1

                    # Identify the smallest item at the current index
                    for j in range(len(pclists)):
                        if pclists[j][i] < smallest_item:
                            smallest_item = pclists[j][i]

                    # Remove all lists with larger items at the current index
                    j = 0
                    while j < len(pclists):
                        if pclists[j][i] > smallest_item:
                            del pclists[j]
                        else:
                            j += 1
                else:
                    break
        return pclists[0]


class SetClass24:
    """
    Represents a pc-set-class
    """
    NUM_PC = 24

    def __init__(self, pcset=None):
        """
        Creates a SetClass
        :param pcset: A pcset to initialize the SetClass
        """
        self._dsym = SetClass24.NUM_PC * 2
        self._ic_vector = [0 for i in range(SetClass24.NUM_PC // 2)]
        self._ic_vector_long = [0 for i in range(SetClass24.NUM_PC // 2 + 1)]
        self._name_prime = ""
        self._pcset = set()
        self._weight_right = True
        if pcset is not None:
            self.pcset = pcset

    def __eq__(self, other):
        return self.pcset == other.pcset

    def __hash__(self):
        return hash(self.name_prime)

    def __len__(self):
        return len(self._pcset)

    def __lt__(self, other):
        if len(self) < len(other) or (len(self) == len(other) and self.name_prime < other.name_prime):
            return True
        return False

    def __ne__(self, other):
        return self.pcset != other.pcset

    def __repr__(self):
        # return "<pctheory.pcset.SetClass24 object at " + str(id(self)) + ">: " + repr(self._pcset)
        return self._name_prime

    def __str__(self):
        return self._name_prime

    @property
    def dsym(self):
        """
        Gets the degree of symmetry of the set-class.
        :return: The degree of symmetry
        """
        return self._dsym

    @property
    def ic_vector(self):
        """
        Gets the IC vector
        :return: The IC vector
        """
        return self._ic_vector

    @property
    def ic_vector_long(self):
        """
        Gets the IC vector in long format
        :return: The IC vector in long format
        """
        return self._ic_vector_long

    @property
    def ic_vector_str(self):
        """
        Gets the IC vector as a string
        :return: The IC vector
        """
        s = "["
        for a in range(len(self._ic_vector) - 1):
            s += f"{self._ic_vector[a]}, "
        s += str(self._ic_vector[len(self._ic_vector) - 1])
        s += "]"
        return s

    @property
    def ic_vector_long_str(self):
        """
        Gets the IC vector in long format as a string
        :return: The IC vector in long format
        """
        s = "["
        for a in range(len(self._ic_vector_long) - 1):
            s += f"{self._ic_vector_long[a]}, "
        s += str(self._ic_vector_long[len(self._ic_vector_long) - 1])
        s += "]"
        return s

    @property
    def name_prime(self):
        """
        Generates the prime-form name (Rahn) for a set-class
        :return: The prime-form name
        """
        return self._name_prime

    @property
    def pcset(self):
        """
        Gets the pcset prime form
        :return: The pcset prime form
        """
        return self._pcset

    @pcset.setter
    def pcset(self, value):
        """
        Updates the pcset prime form
        :param value: The new pcset
        :return:
        """
        self._pcset = SetClass24.calculate_prime_form(value, self._weight_right)
        self._make_names()

    @property
    def weight_right(self):
        """
        Whether or not to weight from the right
        :return: A Boolean
        """
        return self._weight_right

    @weight_right.setter
    def weight_right(self, value: bool):
        """
        Whether or not to weight from the right
        :param value: A Boolean
        :return:
        """
        self._weight_right = value
        self._pcset = SetClass24.calculate_prime_form(self._pcset, self._weight_right)

    @staticmethod
    def calculate_prime_form(pcset: set, weight_from_right: bool = True):
        """
        Calculates the prime form of a pcset
        :param pcset: The pcset
        :param weight_from_right: Whether or not to pack from the right
        :return: The prime form
        """
        prime_set = set()
        if len(pcset) > 0:
            lists_to_weight = []
            int_set = SetClass24._make_int_set(pcset)
            pclist = list(int_set)
            inverted = list(SetClass24._make_int_set(set(SetClass24._invert(pclist))))
            prime_list = None

            # Add regular forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i].append(pclist[i2])
                for i2 in range(0, i):
                    lists_to_weight[i].append(pclist[i2])
                initial_pitch = lists_to_weight[i][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i][i2] -= initial_pitch
                    if lists_to_weight[i][i2] < 0:
                        lists_to_weight[i][i2] += SetClass24.NUM_PC
                lists_to_weight[i].sort()

            # Add inverted forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                for i2 in range(0, i):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                initial_pitch = lists_to_weight[i + len(pclist)][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i + len(pclist)][i2] -= initial_pitch
                    if lists_to_weight[i + len(pclist)][i2] < 0:
                        lists_to_weight[i + len(pclist)][i2] += SetClass24.NUM_PC
                lists_to_weight[i + len(pclist)].sort()

            # Weight lists
            if weight_from_right:
                prime_list = SetClass24._weight_from_right(lists_to_weight)
            else:
                prime_list = SetClass24._weight_left(lists_to_weight)

            # Create pcset
            for pc in prime_list:
                prime_set.add(pitch.PitchClass24(pc))

        return prime_set

    def contains_abstract_subset(self, sc):
        """
        Determines if a set-class is an abstract subset of this set-class
        :param sc: A set-class
        :return: A boolean
        """
        tr = []
        inv = invert(sc.pcset)
        for i in range(SetClass24.NUM_PC):
            tr.append(transpose(sc.pcset, i))
            tr.append(transpose(inv, i))
        for pcs in tr:
            if pcs.issubset(self.pcset):
                return True
        return False

    def get_abstract_complement(self):
        """
        Gets the abstract complement of the SetClass
        :return: The abstract complement SetClass
        """
        csc = SetClass24()
        csc.pcset = get_complement(self._pcset)
        return csc

    def get_abstract_subset_classes(self):
        """
        Gets a set of subset-classes contained in this SetClass
        :return:
        """
        sub = subsets(self._pcset)
        sc = SetClass24()
        subset_classes = {}
        for s in sub:
            sc.pcset = s
            if sc.name_prime not in subset_classes:
                subset_classes[sc.name_prime] = SetClass24(s)
        subset_classes_set = set()
        for s in subset_classes:
            subset_classes_set.add(subset_classes[s])
        return subset_classes_set

    def get_partition_subset_classes(self):
        """
        Gets a set of set-class partitions of this SetClass
        :return:
        """
        p = partitions2(self._pcset)
        p_set = set()
        for part in p:
            p_set.add((SetClass24(part[0]), SetClass24(part[1])))
        return p_set

    @staticmethod
    def _invert(pcseg: list):
        """
        Inverts a pcseg
        :param pcset: The pcseg
        :return: The inverted pcseg
        """
        pcseg2 = []
        for pc in pcseg:
            pcseg2.append(pitch.PitchClass24((pc * (SetClass24.NUM_PC - 1)) % SetClass24.NUM_PC))
        return pcseg2

    @staticmethod
    def _make_int_set(pcset: set):
        """
        Makes an int set
        :param pcset: A pcset
        :return: An int set
        """
        pcset2 = set()
        for pc in pcset:
            pcset2.add(pc.pc)
        return pcset2

    def _make_names(self):
        """
        Makes the names for the set-class
        :return:
        """
        pcseg = list(self._pcset)
        pcseg.sort()
        self._name_prime = str(pcseg)
        self._ic_vector_long = [0 for i in range(13)]
        for pc in self._pcset:
            for pc2 in self._pcset:
                interval = (pc2.pc - pc.pc) % SetClass24.NUM_PC
                if interval > SetClass24.NUM_PC // 2:
                    interval = interval * -1 + SetClass24.NUM_PC
                self._ic_vector_long[interval] += 1
        for i in range(1, len(self._ic_vector_long)):
            self._ic_vector_long[i] //= 2
        self._ic_vector = self._ic_vector_long[1:]
        if len(self._pcset) > 0:
            c = get_corpus(self._pcset)
            self._dsym = (SetClass24.NUM_PC * 2) // len(c)
        else:
            self._dsym = (SetClass24.NUM_PC * 2)

    @staticmethod
    def _weight_from_right(pclists: list):
        """
        Weights pclists from the right
        :param pclists: Pclists
        :return: The most weighted form
        """
        for i in range(len(pclists[0]) - 1, -1, -1):
            if len(pclists) > 1:
                # The smallest item at the current index
                smallest_item = SetClass24.NUM_PC - 1

                # Identify the smallest item at the current index
                for j in range(len(pclists)):
                    if pclists[j][i] < smallest_item:
                        smallest_item = pclists[j][i]

                # Remove all lists with larger items at the current index
                j = 0
                while j < len(pclists):
                    if pclists[j][i] > smallest_item:
                        del pclists[j]
                    else:
                        j += 1

            else:
                break
        return pclists[0]

    @staticmethod
    def _weight_left(pclists: list):
        """
        Weights pclists left
        :param pclists: Pclists
        :return: The most weighted form
        """
        if len(pclists) > 1:
            # The smallest item at the current index
            smallest_item = SetClass24.NUM_PC - 1

            # Identify the smallest item at the last index
            for j in range(0, len(pclists)):
                if pclists[j][len(pclists[0]) - 1] < smallest_item:
                    smallest_item = pclists[j][len(pclists[0]) - 1]

            # Remove all lists with larger items at the current index
            j = 0
            while j < len(pclists):
                if pclists[j][len(pclists[0]) - 1] > smallest_item:
                    del pclists[j]
                else:
                    j += 1

            # Continue processing, but now pack from the left
            for i in range(0, len(pclists[0])):
                if len(pclists) > 1:
                    smallest_item = SetClass24.NUM_PC - 1

                    # Identify the smallest item at the current index
                    for j in range(len(pclists)):
                        if pclists[j][i] < smallest_item:
                            smallest_item = pclists[j][i]

                    # Remove all lists with larger items at the current index
                    j = 0
                    while j < len(pclists):
                        if pclists[j][i] > smallest_item:
                            del pclists[j]
                        else:
                            j += 1
                else:
                    break
        return pclists[0]


def get_complement(pcset: set):
    """
    Gets the complement of a pcset
    :param pcset: A pcset
    :return: The complement pcset
    """
    universal = set()
    if len(pcset) > 0:
        t = type(next(iter(pcset)))
        if t == pitch.PitchClass12:
            for i in range(12):
                universal.add(pitch.PitchClass12(i))
        else:
            for i in range(24):
                universal.add(pitch.PitchClass24(i))
    return universal - pcset


def get_corpus(pcset: set):
    """
    Gets all transformations of a provided pcset
    :param pcset: A pcset
    :return: A set of all transformations of the pcset
    """
    uto = transformations.get_utos24()
    pcsets = set()
    if len(pcset) > 0:
        n = 12
        t = type(next(iter(pcset)))
        if t == pitch.PitchClass24:
            n = 24
        for i in range(n):
            pcsets.add(frozenset(uto[f"T{i}"].transform(pcset)))
            pcsets.add(frozenset(uto[f"T{i}M11"].transform(pcset)))
    return pcsets


def get_uto(original_pcset: set, transformed_pcset: set):
    """
    Finds all UTOs that produce a set that contains transformed_pcset as a proper or improper subset.
    If the list of UTOs is empty, transformed_pcset is not an abstract subset of original_pcset.
    :param original_pcset: The original pcset
    :param transformed_pcset: The new pcset
    :return: A list of UTOs
    """
    utos = None
    utos_final = {}
    if len(original_pcset) == len(transformed_pcset) == 0:
        return utos
    else:
        t = type(next(iter(original_pcset)))
        if t == pitch.PitchClass12:
            utos = transformations.get_utos12()
        else:
            utos = transformations.get_utos24()
        for u in utos:
            if transformed_pcset.issubset(utos[u].transform(original_pcset)):
                utos_final[u] = utos[u]
        return utos_final


def invert(pcset: set):
    """
    Inverts a pcset
    :param pcset: The pcset
    :return: The inverted pcset
    """
    pcset2 = set()
    if len(pcset) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcset)))
        for pc in pcset:
            pcset2.add(t(pc.pc * -1))
    return pcset2


def make_pcset12(*args):
    """
    Makes a pcset
    :param *args: Pcs
    :return: A pcset
    """
    if type(args[0]) == list:
        args = args[0]
    return {pitch.PitchClass24(pc) for pc in args}


def make_pcset24(*args):
    """
    Makes a pcset
    :param *args: Pcs
    :return: A pcset
    """
    if type(args[0]) == list:
        args = args[0]
    return {pitch.PitchClass24(pc) for pc in args}


def make_subset_graph(set_class, smallest_cardinality=1, show_graph=False, size=(800, 1100)):
    """
    Makes a subset graph
    :param set_class: A set-class
    :param smallest_cardinality: The smallest cardinality to include in the graph
    :param show_graph: Whether or not to generate a visualization of the graph
    :param size: The size of the visualized graph
    :return: A graph
    """
    g = networkx.DiGraph()
    scs = list(set_class.get_abstract_subset_classes())
    for s in scs:
        if len(s.pcset) >= smallest_cardinality:
            g.add_node(s.name_prime)
    for i in range(0, len(scs)):
        for j in range(0, i):
            if scs[i].contains_abstract_subset(scs[j]) and len(scs[j].pcset) >= smallest_cardinality:
                g.add_edge(scs[i].name_prime, scs[j].name_prime)
        for j in range(i + 1, len(scs)):
            if scs[i].contains_abstract_subset(scs[j]) and len(scs[j].pcset) >= smallest_cardinality:
                g.add_edge(scs[i].name_prime, scs[j].name_prime)
    if show_graph:
        net = pyvis.network.Network(f"{size[0]}px", f"{size[1]}px", directed=True, bgcolor="#eeeeee",
                                    font_color="#333333", heading="Subset Graph")
        net.toggle_hide_edges_on_drag(False)
        net.barnes_hut()
        net.from_nx(g, default_node_size=40)
        for node in net.nodes:
            node["title"] = node["id"]
            node["color"] = "#41b535"
        for edge in net.edges:
            edge["color"] = "#1c4219"
        net.show("subset_graph.html")
    return g


def multiply(pcset: set, n: int):
    """
    Multiplies a pcset
    :param pcset: The pcset
    :param n: The multiplier
    :return: The multiplied pcset
    """
    pcset2 = set()
    if len(pcset) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcset)))
        for pc in pcset:
            pcset2.add(t(pc.pc * n))
    return pcset2


def partitions2(pcset: set):
    """
    Gets all partitions of a pcset (size 2 or 1)
    :param pcset: A pcset
    :return: A list of all partitions
    """
    subs = subsets(pcset)
    partitions_dict = {}
    partitions_list = []
    len_pcset = (len(pcset) + 1) // 2 if len(pcset) % 2 else len(pcset) // 2
    for sub in subs:
        if len(sub) <= len_pcset:
            s = frozenset(sub)
            d = frozenset(pcset.difference(sub))
            if d not in partitions_dict:
                partitions_dict[s] = d
    for s in partitions_dict:
        partitions_list.append((set(s), set(partitions_dict[s])))
    return partitions_list


def set_class_filter12(name: str, sets: list):
    """
    Filters a list of pcsets
    :param name: The name to find
    :param sets: A list of sets to filter
    :return: A filtered list
    """
    newlist = []
    sc = SetClass12()
    for s in sets:
        sc.pcset = s
        if sc.name_prime == name or sc.name_forte == name or sc.name_morris == name:
            newlist.append(s)
    return newlist


def subsets(pcset: set):
    """
    Gets all subsets of a pcset, using the bit masking solution from
    https://afteracademy.com/blog/print-all-subsets-of-a-given-set
    :param pcset: A pcset
    :return: A list containing all subsets of the pcset
    """
    total = 2 ** len(pcset)
    sub = []
    if total > 1:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcset)))
        pcseg = list(pcset)
        pcseg.sort()
        for index in range(total):
            sub.append([])
            for i in range(len(pcset)):
                if index & (1 << i):
                    sub[index].append(t(pcseg[i].pc))
        sub.sort()
    return sub


def transpose(pcset: set, n: int):
    """
    Transposes a pcset
    :param pcset: The pcset
    :param n: The index of transposition
    :return: The transposed pcset
    """
    pcset2 = set()
    if len(pcset) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcset)))
        for pc in pcset:
            pcset2.add(t(pc.pc + n))
    return pcset2


def transpositional_combination(pcset1: set, pcset2: set):
    """
    Transpositionally combines (TC) two pcsets. This is Boulez's "multiplication."
    :param pcset1: A pcset
    :param pcset2: A pcset
    :return: The TC pcset
    """
    pcset3 = set()
    if len(pcset1) > 0 and len(pcset2) > 0:
        # Need to support both PitchClass12 and PitchClass24, so use a type alias
        t = type(next(iter(pcset1)))
        for pc2 in pcset2:
            for pc1 in pcset1:
                pcset3.add(t(pc1.pc + pc2.pc))
    return pcset3


def visualize(pcset: set):
    """
    Visualizes a pcset
    :param pcset: A pcset
    :return: A visualization
    """
    line = ""
    if len(pcset) > 0:
        t = type(next(iter(pcset)))
        n = 12
        if t == pitch.PitchClass24:
            n = 24
        for i in range(n):
            if t(i) in pcset:
                line += "X"
            else:
                line += " "
    return line



"""
File: pitch.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class PitchClass12:
    def __init__(self, pc: int = 0):
        """
        Creates a PitchClass
        :param pc: The pitch class integer
        """
        self._pc = pc % 12

    def __add__(self, other):
        return PitchClass12(self.pc + other.pc)

    def __eq__(self, other):
        return self.pc == other.pc

    def __ge__(self, other):
        return self.pc >= other.pc

    def __gt__(self, other):
        return self.pc > other.pc

    def __hash__(self):
        return self.pc

    def __le__(self, other):
        return self.pc <= other.pc

    def __lt__(self, other):
        return self.pc < other.pc

    def __mul__(self, other):
        return PitchClass12(self.pc * other.pc)

    def __ne__(self, other):
        return self.pc != other.pc

    def __repr__(self):
        # return "<pctheory.pitch.PitchClass object at " + str(id(self)) + ">: " + self.pc_char
        return self.pc_str

    def __str__(self):
        return self.pc_str

    def __sub__(self, other):
        return PitchClass12(self.pc - other.pc)

    @property
    def pc(self):
        """
        The pitch-class integer
        :return: The pitch-class integer
        """
        return self._pc

    @pc.setter
    def pc(self, value: int):
        """
        The pitch-class integer
        :param value: The new pitch-class integer
        :return:
        """
        self._pc = value % 12
        if self._pc < 0:
            self._pc += 12

    @property
    def pc_str(self):
        """
        The pitch-class string
        :return: The pitch-class string
        """
        if self._pc == 10:
            return 'A'
        elif self._pc == 11:
            return 'B'
        else:
            return str(self._pc)


class PitchClass24:
    def __init__(self, pc: int = 0):
        """
        Creates a quarter-tone PitchClass
        :param pc: The pitch class integer
        """
        self._pc = pc % 24

    def __add__(self, other):
        return PitchClass24(self.pc + other.pc)

    def __eq__(self, other):
        return self.pc == other.pc

    def __ge__(self, other):
        return self.pc >= other.pc

    def __gt__(self, other):
        return self.pc > other.pc

    def __hash__(self):
        return self.pc

    def __le__(self, other):
        return self.pc <= other.pc

    def __lt__(self, other):
        return self.pc < other.pc

    def __mul__(self, other):
        return PitchClass24(self.pc * other.pc)

    def __ne__(self, other):
        return self.pc != other.pc

    def __repr__(self):
        # return "<pctheory.pitch.PitchClass24 object at " + str(id(self)) + ">: " + self.pc_char
        return self.pc_str

    def __str__(self):
        return self.pc_str

    def __sub__(self, other):
        return PitchClass24(self.pc - other.pc)

    @property
    def pc(self):
        """
        The pitch-class integer
        :return: The pitch-class integer
        """
        return self._pc

    @pc.setter
    def pc(self, value: int):
        """
        The pitch-class integer
        :param value: The new pitch-class integer
        :return:
        """
        self._pc = value % 24
        if self._pc < 0:
            self._pc += 24

    @property
    def pc_str(self):
        """
        The pitch-class string
        :return: The pitch-class string
        """
        if self._pc < 10:
            return f"0{self._pc}"
        else:
            return str(self._pc)


class Pitch12(PitchClass12):
    """
    Represents a pitch
    """
    def __init__(self, p: int = 0, pname: str = 0):
        """
        Creates a Pitch
        :param p: The pitch integer
        :param pname: The pitch name as a string
        """
        pc = p % 12
        super().__init__(pc)
        self._p = p
        self._pname = pname
        self.pc = p

    def __add__(self, other):
        return Pitch12(self.p + other.p)

    def __eq__(self, other):
        return self.p == other.p

    def __ge__(self, other):
        return self.p >= other.p

    def __gt__(self, other):
        return self.p > other.p

    def __hash__(self):
        return self.p

    def __le__(self, other):
        return self.p <= other.p

    def __lt__(self, other):
        return self.p < other.p

    def __mul__(self, other):
        return Pitch12(self.p * other.p)

    def __ne__(self, other):
        return self.p != other.p

    def __repr__(self):
        # return "<pctheory.pitch.Pitch object at " + str(id(self)) + ">: " + str(self._p)
        return str(self._p)

    def __str__(self):
        return str(self._p)

    def __sub__(self, other):
        return Pitch12(self.p - other.p)

    @property
    def p(self):
        """
        The pitch integer
        :return: The pitch integer
        """
        return self._p

    @p.setter
    def p(self, value: int):
        """
        The pitch integer
        :param value: The new pitch integer
        :return:
        """
        self._p = value
        self.pc = self._p

    @property
    def pname(self):
        """
        The pitch name of the pitch
        :return: The pitch name
        """
        return self._pname

    @pname.setter
    def pname(self, value: str):
        """
        The pitch name of the pitch
        :param value: The new name
        :return:
        """
        self._pname = value

    @property
    def midi(self):
        """
        The MIDI version of the pitch integer
        :return: The MIDI version of the pitch integer
        """
        return self._p + 60

    @midi.setter
    def midi(self, value):
        """
        The MIDI version of the pitch integer
        :param value: The new MIDI version of the pitch integer
        :return:
        """
        self._p = value - 60
        self.pc = self._p


class Pitch24(PitchClass24):
    """
    Represents a pitch
    """
    def __init__(self, p: int = 0, pname: str = 0):
        """
        Creates a Pitch
        :param p: The pitch integer
        :param pname: The pitch name as a string
        """
        pc = p % 24
        super().__init__(pc)
        self._p = p
        self._pname = pname
        self.pc = p

    def __add__(self, other):
        return Pitch24(self.p + other.p)

    def __eq__(self, other):
        return self.p == other.p

    def __ge__(self, other):
        return self.p >= other.p

    def __gt__(self, other):
        return self.p > other.p

    def __hash__(self):
        return self.p

    def __le__(self, other):
        return self.p <= other.p

    def __lt__(self, other):
        return self.p < other.p

    def __mul__(self, other):
        return Pitch24(self.p * other.p)

    def __ne__(self, other):
        return self.p != other.p

    def __repr__(self):
        # return "<pctheory.pitch.Pitch object at " + str(id(self)) + ">: " + str(self._p)
        return str(self._p)

    def __str__(self):
        return str(self._p)

    def __sub__(self, other):
        return Pitch24(self.p - other.p)

    @property
    def p(self):
        """
        The pitch integer
        :return: The pitch integer
        """
        return self._p

    @p.setter
    def p(self, value: int):
        """
        The pitch integer
        :param value: The new pitch integer
        :return:
        """
        self._p = value
        self.pc = self._p

    @property
    def pname(self):
        """
        The pitch name of the pitch
        :return: The pitch name
        """
        return self._pname

    @pname.setter
    def pname(self, value: str):
        """
        The pitch name of the pitch
        :param value: The new name
        :return:
        """
        self._pname = value
